package com.tzld.piaoquan.content.understanding.service.impl;

import com.alibaba.fastjson.JSON;
import com.aliyun.odps.data.Record;
import com.aliyun.oss.model.ObjectMetadata;
import com.aliyun.oss.model.StorageClass;
import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.base.Stopwatch;
import com.tzld.commons.aliyun.oss.AliyunOssManager;
import com.tzld.piaoquan.content.understanding.common.base.Constant;
import com.tzld.piaoquan.content.understanding.common.enums.ContentTypeEnum;
import com.tzld.piaoquan.content.understanding.common.enums.DataStatusEnum;
import com.tzld.piaoquan.content.understanding.common.enums.ExceptionEnum;
import com.tzld.piaoquan.content.understanding.common.exception.CommonException;
import com.tzld.piaoquan.content.understanding.dao.mapper.*;
import com.tzld.piaoquan.content.understanding.model.dto.ContentDTO;
import com.tzld.piaoquan.content.understanding.model.dto.ContentUnderstandingResultDTO;
import com.tzld.piaoquan.content.understanding.model.dto.TaskResultDTO;
import com.tzld.piaoquan.content.understanding.model.param.*;
import com.tzld.piaoquan.content.understanding.model.po.*;
import com.tzld.piaoquan.content.understanding.model.vo.ContentUnderstandingResultVO;
import com.tzld.piaoquan.content.understanding.model.vo.TaskResultVO;
import com.tzld.piaoquan.content.understanding.rocketmq.producer.OldVersionRemainVideoUnderstandingProducer;
import com.tzld.piaoquan.content.understanding.service.ContentService;
import com.tzld.piaoquan.content.understanding.service.LoghubService;
import com.tzld.piaoquan.content.understanding.service.ODPSManager;
import com.tzld.piaoquan.content.understanding.service.PipelineService;
import com.tzld.piaoquan.content.understanding.util.HttpClientUtil;
import com.tzld.piaoquan.content.understanding.util.HttpPoolClient;
import com.tzld.piaoquan.content.understanding.util.RetryUtil;
import com.tzld.piaoquan.content.understanding.util.UrlUtil;
import com.xxl.job.core.log.XxlJobLogger;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.apis.message.Message;
import org.apache.rocketmq.client.apis.message.MessageId;
import org.apache.rocketmq.client.apis.producer.SendReceipt;
import org.apache.rocketmq.client.core.RocketMQClientTemplate;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.time.Duration;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @author supeng
 */
@Slf4j
@Service
public class ContentServiceImpl implements ContentService {

    @Autowired
    private LoghubService loghubService;

    @Autowired
    private PipelineService pipelineService;

    @Value("${video.analyse.pipelineId:1}")
    private Long videoAnalysePipelineId;

    @ApolloJsonValue("${content.understand.pipeline.config:{}}")
    private Map<String, Long> configMap;
    @Autowired
    private CuPipelineMapper cuPipelineMapper;
    @Autowired
    private CuTaskMapper cuTaskMapper;
    @Autowired
    private CuVideoUnderstandingResultMapper cuVideoUnderstandingResultMapper;

    private static final Integer MAX_SIZE = 20;

    @Autowired
    private ODPSManager odpsManager;

    @Autowired
    private WxVideoMapper wxVideoMapper;

    @Autowired
    private WxVideoTranscodeMapper wxVideoTranscodeMapper;

    private static final String TRANSCODE_CHANNEL = "WEIZHEN";

    @Value("${video.new.recommend.table:content_ai_tag_recommend}")
    private String newRecommendTable;

    @Value("${video.top.return.table:content_ai_tag_return_top_merge}")
    private String topReturnTable;

    @Value("${video.top.history.table:history_top2000_add_aitags}")
    private String topHistoryTable;

    @Value("${video.analyse:overwrite:0}")
    private String overwriteAnalyse;

    @Value("${old.version.video.understanding.pipelineId:0}")
    private Long oldVersionPipelineId;

    @Autowired
    private AliyunOssManager aliyunOssManager;

    public static final String TRANSCODE_BUCKET = "art-pubbucket";

    private static final Integer EXPIRE_DAYS = 7;

    @Autowired
    private OldVersionRemainVideoUnderstandingProducer oldVersionRemainVideoUnderstandingProducer;

    @Override
    public void analyse(ContentAnalyseParam param) {
        if (Objects.isNull(param) || Objects.isNull(param.getType())) {
            throw new CommonException(ExceptionEnum.PARAMS_INVALID);
        }
        param.setPipelineId(videoAnalysePipelineId);
        pipelineService.execute(param);
    }

    @Override
    public void understand(ContentUnderstandParam param) {
//        if (Objects.isNull(param) || Objects.isNull(param.getContentTypes()) || param.getContentTypes().isEmpty()
//                || Objects.isNull(param.getContentInfo())) {
//            throw new CommonException(ExceptionEnum.PARAMS_INVALID);
//        }
//        //生成jobId
//        String jobId = UUID.randomUUID().toString().replace("-", "") + System.currentTimeMillis();
//        for (Integer contentType : param.getContentTypes()) {
//            Long pipelineId = configMap.get(contentType.toString());
//            if (Objects.isNull(pipelineId)) {
//                //TODO 日志上报 && 记录状态
//                log.info("jobId = {}", jobId);
//                continue;
//            }
//            ContentUnderstandDTO dto = new ContentUnderstandDTO();
//            BeanUtils.copyProperties(param.getContentInfo(), dto);
//            dto.setPipelineId(pipelineId);
//            dto.setContentType(contentType);
//            pipelineService.execute(dto);
//        }
    }

    @Override
    public TaskResultVO submitTasks(SubmitTasksParam param) {
        //参数校验
        if (Objects.isNull(param) || Objects.isNull(param.getContents()) || param.getContents().isEmpty()) {
            throw new CommonException(ExceptionEnum.PARAMS_INVALID);
        }
        List<ContentDTO> contentDTOList = param.getContents();
        //数据校验
        for (ContentDTO contentDTO : contentDTOList) {
            if (Objects.isNull(contentDTO) || Objects.isNull(contentDTO.getPipelineId())
                    || Objects.isNull(contentDTO.getContent()) || Objects.isNull(contentDTO.getVideoId())) {
                throw new CommonException(ExceptionEnum.PARAMS_INVALID, "videoId、pipelineId、content参数不能为空");
            }
            Long pipelineId = contentDTO.getPipelineId();
            CuPipelineExample example = new CuPipelineExample();
            example.createCriteria().andIdEqualTo(pipelineId).andDataStatusEqualTo(DataStatusEnum.VALID.getValue());
            List<CuPipeline> list = cuPipelineMapper.selectByExample(example);
            if (Objects.isNull(list) || list.isEmpty()) {
                throw new CommonException(ExceptionEnum.PARAMS_INVALID, "pipelineId [" + pipelineId + "]无效");
            }
            CuPipeline cuPipeline = list.get(0);
            if (Objects.isNull(cuPipeline) || !ContentTypeEnum.containsValue(cuPipeline.getContentType())) {
                throw new CommonException(ExceptionEnum.PARAMS_INVALID, "pipelineId [" + pipelineId + "]配置异常");
            }
            contentDTO.setContentType(cuPipeline.getContentType());
        }
        //执行任务
        TaskResultVO taskResultVO = new TaskResultVO();
        List<TaskResultDTO> taskResultDTOS = new ArrayList<>();
        for (ContentDTO contentDTO : contentDTOList) {
            TaskResultDTO taskResultDTO = new TaskResultDTO();
            BeanUtils.copyProperties(contentDTO, taskResultDTO);
            String taskId = pipelineService.executeTask(contentDTO);
            taskResultDTO.setTaskId(taskId);
            taskResultDTOS.add(taskResultDTO);
        }
        taskResultVO.setTaskResults(taskResultDTOS);
        return taskResultVO;
    }

    @Override
    public TaskResultVO queryTasks(QueryTasksParam param) {
        //参数校验
        if (Objects.isNull(param) || Objects.isNull(param.getTaskIds()) || param.getTaskIds().isEmpty()) {
            throw new CommonException(ExceptionEnum.PARAMS_INVALID);
        }
        CuTaskExample example = new CuTaskExample();
        example.createCriteria().andTaskIdIn(param.getTaskIds());
        List<CuTask> list = cuTaskMapper.selectByExample(example);
        TaskResultVO taskResultVO = new TaskResultVO();
        if (Objects.isNull(list) || list.isEmpty()) {
            return taskResultVO;
        }
        List<TaskResultDTO> taskResultDTOS = new ArrayList<>();
        for (CuTask task : list) {
            TaskResultDTO taskResultDTO = new TaskResultDTO();
            BeanUtils.copyProperties(task, taskResultDTO);
            taskResultDTOS.add(taskResultDTO);
        }
        taskResultVO.setTaskResults(taskResultDTOS);
        return taskResultVO;
    }

    @Override
    public ContentUnderstandingResultVO getVideoUnderstandingResult(GetVideoUnderstandingResultParam param) {
        if (Objects.isNull(param) || Objects.isNull(param.getContents()) || param.getContents().isEmpty()) {
            throw new CommonException(ExceptionEnum.PARAMS_INVALID, "参数不能为空");
        }
        if (param.getContents().size() > MAX_SIZE) {
            throw new CommonException(ExceptionEnum.PARAMS_INVALID, "查询结果数不能超过" + MAX_SIZE + "条");
        }
        ContentUnderstandingResultVO contentUnderstandingResultVO = new ContentUnderstandingResultVO();
        List<ContentUnderstandingResultDTO> contentUnderstandingResultDTOS = new ArrayList<>();
        for (ContentDTO contentDTO : param.getContents()) {
            if (Objects.isNull(contentDTO) || Objects.isNull(contentDTO.getPipelineId())
                    || Objects.isNull(contentDTO.getContent()) || Objects.isNull(contentDTO.getVideoId())) {
                continue;
            }
            CuVideoUnderstandingResultExample example = new CuVideoUnderstandingResultExample();
            example.createCriteria().andVideoIdEqualTo(contentDTO.getVideoId()).andPipelineIdEqualTo(contentDTO.getPipelineId());
            List<CuVideoUnderstandingResult> list = cuVideoUnderstandingResultMapper.selectByExample(example);
            if (Objects.isNull(list) || list.isEmpty()) {
                continue;
            }
            ContentUnderstandingResultDTO contentUnderstandingResultDTO = new ContentUnderstandingResultDTO();
            BeanUtils.copyProperties(contentDTO, contentUnderstandingResultDTO);
            contentUnderstandingResultDTOS.add(contentUnderstandingResultDTO);
        }
        contentUnderstandingResultVO.setResults(contentUnderstandingResultDTOS);
        return contentUnderstandingResultVO;
    }

    @Override
    public void newRecommendVideoUnderstandingHandler() {
        Stopwatch stopwatch = Stopwatch.createStarted();
        List<Record> newRecommendRecords = odpsManager.tableTunnelQuery("loghubods", newRecommendTable);
        if (Objects.nonNull(newRecommendRecords) && !newRecommendRecords.isEmpty()) {
            XxlJobLogger.log("table = {} records size = {}", newRecommendTable, newRecommendRecords.size());
            for (Record record : newRecommendRecords) {
                try {
                    Long videoId = record.getBigint("videoid");
                    if (Objects.isNull(videoId)) {
                        XxlJobLogger.log("table = {} videoId 为 null", newRecommendTable);
                        continue;
                    }
                    String videoUrl = getVideoUrl(videoId);
                    if (Objects.isNull(videoUrl)) {
                        XxlJobLogger.log("table = {} videoId = {} videoUrl is null", newRecommendTable, videoId);
                        continue;
                    }
                    SubmitTasksParam submitTasksParam = new SubmitTasksParam();
                    List<ContentDTO> contentDTOList = new ArrayList<>();
                    ContentDTO contentDTO = new ContentDTO();
                    contentDTO.setVideoId(videoId);
                    contentDTO.setContentType(ContentTypeEnum.VIDEO.getValue());
                    contentDTO.setContent(videoUrl);
                    contentDTO.setPipelineId(oldVersionPipelineId);
                    contentDTOList.add(contentDTO);
                    submitTasksParam.setContents(contentDTOList);
                    submitTasks(submitTasksParam);
                } catch (Exception e) {
                    XxlJobLogger.log("record = {} error", record);
                }
            }
        } else {
            XxlJobLogger.log("table = {} records is empty", newRecommendTable);
        }
        XxlJobLogger.log("table = {} execute time = {}", newRecommendTable, stopwatch.stop().elapsed(TimeUnit.MILLISECONDS));
    }

    @Override
    public void dayTopVideoUnderstandingHandler() {
        Stopwatch stopwatch = Stopwatch.createStarted();
        List<Record> newRecommendRecords = odpsManager.tableTunnelQuery("loghubods", topReturnTable);
        if (Objects.nonNull(newRecommendRecords) && !newRecommendRecords.isEmpty()) {
            XxlJobLogger.log("table = {} records size = {}", topReturnTable, newRecommendRecords.size());
            for (Record record : newRecommendRecords) {
                try {
                    Long videoId = record.getBigint("videoid");
                    if (Objects.isNull(videoId)) {
                        XxlJobLogger.log("table = {} videoId 为 null", topReturnTable);
                        continue;
                    }
                    String videoUrl = getVideoUrl(videoId);
                    if (Objects.isNull(videoUrl)) {
                        XxlJobLogger.log("table = {} videoId = {} videoUrl is null", topReturnTable, videoId);
                        continue;
                    }
                    SubmitTasksParam submitTasksParam = new SubmitTasksParam();
                    List<ContentDTO> contentDTOList = new ArrayList<>();
                    ContentDTO contentDTO = new ContentDTO();
                    contentDTO.setVideoId(videoId);
                    contentDTO.setContentType(ContentTypeEnum.VIDEO.getValue());
                    contentDTO.setContent(videoUrl);
                    contentDTO.setPipelineId(oldVersionPipelineId);
                    contentDTOList.add(contentDTO);
                    submitTasksParam.setContents(contentDTOList);
                    TaskResultVO taskResultVO = submitTasks(submitTasksParam);
                } catch (Exception e) {
                    XxlJobLogger.log("record = {} error", record);
                }
            }
        } else {
            XxlJobLogger.log("table = {} records is empty", topReturnTable);
        }
        XxlJobLogger.log("table = {} execute time = {}", topReturnTable, stopwatch.stop().elapsed(TimeUnit.MILLISECONDS));
    }

    /**
     * 获取视频URL
     * <p>
     * 有多码率文件使用多码率文件（多码率文件目前不会被冻结/冷归档/深度冷归档和删除）
     * 没有多码率文件使用转码文件，转码文件被冻结/冷归档/深度冷归档，则需要解冻
     * 没有转码文件使用原文件，原文件被冻结/冷归档/深度冷归档，则需要解冻
     * 没有原文件则失败
     *
     * @param videoId
     * @return
     */
    private String getVideoUrl(Long videoId) {
        //获取多码率
        String videoUrl = null;
        WxVideoTranscodeExample transcodeExample = new WxVideoTranscodeExample();
        transcodeExample.createCriteria().andVideoIdEqualTo(videoId).andTransChannelEqualTo(TRANSCODE_CHANNEL)
                .andTranscodeStatusEqualTo(3);
        List<WxVideoTranscode> transcodeList = wxVideoTranscodeMapper.selectByExample(transcodeExample);
        if (Objects.nonNull(transcodeList) && !transcodeList.isEmpty()) {
            WxVideoTranscode wxVideoTranscode = transcodeList.get(0);
            String transcodeVideoPath = wxVideoTranscode.getTranscodeVideoPath();
            if (Objects.nonNull(transcodeVideoPath)) {
                videoUrl = UrlUtil.processUrl(Constant.MULTI_CDN_HTTP_DOMAIN, transcodeVideoPath);
            }
        }
        //没有多码率文件
        if (Objects.isNull(videoUrl)) {
            WxVideo wxVideo = wxVideoMapper.selectByPrimaryKey(videoId);
            if (Objects.isNull(wxVideo)) {
                XxlJobLogger.log("table = {} videoId = {} wxVideo is null", topReturnTable, videoId);
                return null;
            }
            String transedVideoPath = wxVideo.getTransedVideoPath();
            ObjectMetadata objectMetadata = aliyunOssManager.getObjectMetadata(TRANSCODE_BUCKET, transedVideoPath);
            if (Objects.nonNull(objectMetadata)) {
                boolean isCodeArchive = Objects.equals(objectMetadata.getObjectStorageClass(), StorageClass.ColdArchive);
                boolean isDeepColdArchive = Objects.equals(objectMetadata.getObjectStorageClass(), StorageClass.ColdArchive);
                if (isCodeArchive || isDeepColdArchive) {
                    try {
                        //throw NPE 未解冻 true 解冻完成 false 解冻中
                        if (objectMetadata.isRestoreCompleted()) {
                            //解冻完成
                            videoUrl = UrlUtil.processUrl(Constant.VIDEO_CDN_HTTP_DOMAIN, transedVideoPath);
                        } else {
                            //解冻中
                            return null;
                        }
                    } catch (Exception e) {
                        log.error("{} {} not restoreObject", videoId, videoUrl);
                    }
                    //解冻
                    aliyunOssManager.restoreObject(TRANSCODE_BUCKET, transedVideoPath, EXPIRE_DAYS);
                    Map<String, Object> messageMap = new HashMap<>();
                    messageMap.put("videoId", videoId);
                    Duration duration;
                    if (isCodeArchive) {
                        duration = Duration.ofHours(1);
                    } else {
                        duration = Duration.ofHours(12);
                    }
                    try {
                        MessageId messageId = oldVersionRemainVideoUnderstandingProducer.syncSendDelayMessage(messageMap, duration);
                        log.info("syncSendDelayMessage messageId = {}", messageId);
                    } catch (Exception e) {
                        log.error("syncSendDelayMessage error messageMap = {}", e, messageMap);
                    }
                    return null;
                } else {
                    videoUrl = UrlUtil.processUrl(Constant.VIDEO_CDN_HTTP_DOMAIN, transedVideoPath);
                }
            }
        }
        //没有转码文件
        if (Objects.isNull(videoUrl)) {
            WxVideo wxVideo = wxVideoMapper.selectByPrimaryKey(videoId);
            if (Objects.isNull(wxVideo)) {
                XxlJobLogger.log("table = {} videoId = {} wxVideo is null", topReturnTable, videoId);
                return null;
            }
            String videoPath = wxVideo.getVideoPath();
            ObjectMetadata objectMetadata = aliyunOssManager.getObjectMetadata(TRANSCODE_BUCKET, videoPath);
            if (Objects.nonNull(objectMetadata)) {
                boolean isCodeArchive = Objects.equals(objectMetadata.getObjectStorageClass(), StorageClass.ColdArchive);
                boolean isDeepColdArchive = Objects.equals(objectMetadata.getObjectStorageClass(), StorageClass.ColdArchive);
                if (isCodeArchive || isDeepColdArchive) {
                    try {
                        //throw NPE 未解冻 true 解冻完成 false 解冻中
                        if (objectMetadata.isRestoreCompleted()) {
                            //解冻完成
                            videoUrl = UrlUtil.processUrl(Constant.VIDEO_CDN_HTTP_DOMAIN, videoPath);
                        } else {
                            //解冻中
                            return null;
                        }
                    } catch (Exception e) {
                        log.error("{} {} not restoreObject", videoId, videoUrl);
                    }
                    //解冻
                    aliyunOssManager.restoreObject(TRANSCODE_BUCKET, videoPath, EXPIRE_DAYS);
                    Map<String, Object> messageMap = new HashMap<>();
                    messageMap.put("videoId", videoId);
                    Duration duration;
                    if (isCodeArchive) {
                        duration = Duration.ofHours(1);
                    } else {
                        duration = Duration.ofHours(12);
                    }
                    try {
                        oldVersionRemainVideoUnderstandingProducer.syncSendDelayMessage(messageMap, duration);
                    } catch (Exception e) {
                        log.error("syncSendDelayMessage error messageMap = {}", e, messageMap);
                    }
                    return null;
                } else {
                    videoUrl = UrlUtil.processUrl(Constant.VIDEO_CDN_HTTP_DOMAIN, videoPath);
                }
            }
        }
        return videoUrl;
    }

    @Override
    public void remainVideoUnderstandingHandler() {

    }

    @Override
    public boolean remainVideoUnderstandingHandler(Long videoId) {
        try {
            String videoUrl = getVideoUrl(videoId);
            if (Objects.isNull(videoUrl)) {
                XxlJobLogger.log("table = {} videoId = {} videoUrl is null", newRecommendTable, videoId);
                return false;
            }
            SubmitTasksParam submitTasksParam = new SubmitTasksParam();
            List<ContentDTO> contentDTOList = new ArrayList<>();
            ContentDTO contentDTO = new ContentDTO();
            contentDTO.setVideoId(videoId);
            contentDTO.setContentType(ContentTypeEnum.VIDEO.getValue());
            contentDTO.setContent(videoUrl);
            contentDTO.setPipelineId(oldVersionPipelineId);
            contentDTOList.add(contentDTO);
            submitTasksParam.setContents(contentDTOList);
            TaskResultVO taskResultVO = submitTasks(submitTasksParam);
            if (Objects.nonNull(taskResultVO) && Objects.nonNull(taskResultVO.getTaskResults())
                    && taskResultVO.getTaskResults().size() > 0 && Objects.nonNull(taskResultVO.getTaskResults().get(0))
                    && Objects.nonNull(taskResultVO.getTaskResults().get(0).getTaskId())) {
                return true;
            }
        } catch (Exception e) {
            log.error("remainVideoUnderstandingHandler error", e);
        }
        return false;
    }
}
